# -*- coding: utf-8 -*-


def process_hits(hits, max_score):
    "Process alignment hits by removing overlaps"
    res = []
    for hit in sorted(hits, key=lambda x: (x.RefStart, x.RefEnd)):
        if hit.Score > max_score:
            continue
        if len(res) == 0:
            res.append(hit)
        else:
            last = res[-1]
            if last.RefEnd > hit.RefStart and hit.Score < last.Score:
                # Swap with top element:
                res[-1] = hit
                pass
            else:
                res.append(hit)
    return tuple(res)



import unittest
import os
from os import path
import subprocess


class TestIntegration(unittest.TestCase):

    def testIntegration(self):
        """ Integration test. """
        base = path.dirname(__file__)
        test_base = path.join(base, 'data')

        barcodes = path.join(test_base, 'barcodes.fas')
        input_fasta = path.join(test_base, 'ref.fq.gz')
        output_fasta = path.join(test_base, 'test_output.fq.gz')
        expected_output = path.join(test_base, 'expected_output.fas')

        subprocess.call("{} {} {} {} {}".format('pychopper', "-Y 0 -B 3 -q 0.5 -m edlib -b", barcodes, input_fasta, output_fasta), shell=True)
        retval = subprocess.call(['cmp', output_fasta, expected_output])
        self.assertEqual(retval, 0)
        os.remove(output_fasta)

    def testIntegration_umi(self):
        """ Integration test. """
        base = path.dirname(__file__)
        test_base = path.join(base, 'data')

        input_fasta = path.join(test_base, 'PCS111_umi_test_reads.fastq.gz')
        output_fasta = path.join(test_base, 'test_output_umi.fq.gz')
        expected_output = path.join(test_base, 'PCS111_umi_test_reads_expected.fastq')

        subprocess.call("{} {} {} {}".format('pychopper', "-U -m edlib -k PCS111", input_fasta, output_fasta), shell=True)
        retval = subprocess.call(['cmp', output_fasta, expected_output])
        self.assertEqual(retval, 0)
        os.remove(output_fasta)




# -*- coding: utf-8 -*-

import parasail
import re
from pychopper.common_structures import Hit

# Fixme: allow setting this from the command line?
DEFAULT_ALIGN_PARAMS = {'match': 1,
                        'mismatch': -2,
                        'gap_open': 1,
                        'gap_extend': 1}
DEFAULT_SUBS_MAT = parasail.matrix_create("ACGT", DEFAULT_ALIGN_PARAMS['match'], DEFAULT_ALIGN_PARAMS['mismatch'])
re_split_cigar = re.compile(r"(?P<len>\d+)(?P<op>\D+)")


def first_cigar(cigar):
    """Extract details of the first operation in a cigar string."""
    m = re.search(re_split_cigar, cigar)
    return int(m.group('len')), m.group('op')


def process_alignment(aln, query, query_name, aln_params):
    """ Process an alignment, extracting score, start and end.
    """
    res = {}
    res['query_name'] = query_name
    res['score'] = aln.score
    max_score = float(aln_params['match'] * len(query))
    res['norm_score'] = (max_score - aln.score) / max_score
    fo = first_cigar(aln.cigar.decode.decode())
    res['ref_start'] = aln.cigar.beg_ref
    res['ref_end'] = aln.end_ref + 1
    res['query_start'] = aln.cigar.beg_query
    res['query_end'] = aln.end_query + 1
    if fo[1] == 'I':
        res['query_start'] += fo[0]
    if fo[1] == 'D':
        res['ref_start'] += fo[0]
    return res


def pair_align(reference, query, query_name, subs_mat, params):
    """ Perform pairwise local alignment using parsail-python """
    aln = parasail.sw_trace_striped_32(query, reference, params['gap_open'], params['gap_extend'], subs_mat)
    return process_alignment(aln, query, query_name, params)


def refine_locations(read, all_primers, locations, aln_params=DEFAULT_ALIGN_PARAMS, subs_mat=DEFAULT_SUBS_MAT):
    "Refine alignment edges based on local alignment"
    seq = read.Seq

    proc_locations = []
    for loc in locations:
        aln = pair_align(seq[loc.RefStart:loc.RefEnd], all_primers[loc.Query], loc.Query, subs_mat, aln_params)
        rscore = aln['norm_score']
        rloc = Hit(loc.Ref, loc.RefStart + aln["ref_start"], loc.RefStart + aln["ref_end"], loc.Query, aln["query_start"], aln["query_end"], rscore)
        proc_locations.append(rloc)
    return proc_locations



# -*- coding: utf-8 -*-

import sys
import os
import re
import subprocess as sp
import itertools
from collections import defaultdict
from pychopper.common_structures import Hit
from pychopper import utils


def _parse_hmmscan_tab(lines, reads):
    "Prase nhmmscan tabular output"
    res = []
    for line in lines:
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith('#'):
            continue
        tmp = re.split(r'\s+', line)
        hit = Hit(tmp[2], int(tmp[6]), int(tmp[7]) + 1, tmp[0], int(tmp[4]), int(tmp[5]) + 1, float(tmp[12]))
        res.append(hit)

    buff = defaultdict(list)
    for x, y in itertools.groupby(res, lambda x: x.Ref):
        buff[x] = list(y)
    for r in reads:
        yield buff[r.Id]


def find_locations(reads, phmm_file, E, pool, min_batch):
    "Find alignment hits of all primers in all reads using the pHMM/nhmmscan backend"
    batches = list(utils.batch(reads, min_batch))
    for res in pool.map(_find_locations_single, zip(batches, [(phmm_file, E, 1)] * len(batches))):
        for h in res:
            yield list(h)


def _find_locations_single(params):
    "Find alignment hits of all primers in a single reads using the pHMM/nhmmscan backend"
    reads = params[0]
    phmm_file, E, threads = params[1]
    if not os.path.isfile(phmm_file):
        raise Exception("Profile HMM file is invalid: " + phmm_file)
    cmd = "nhmmscan --notextw --max -E {} --cpu {} --watson -o /dev/null --tblout /dev/stdout {} -"
    cmd = cmd.format(E, threads, phmm_file)
    with sp.Popen(cmd, shell=True, stdin=sp.PIPE, stdout=sp.PIPE) as proc:
        in_data = ""
        for read in reads:
            in_data += ">{}\n{}\n".format(read.Id, read.Seq)
        dout, derr = proc.communicate(in_data.encode())

    if derr is not None:
        print("Failed to run hmmscan with model {} and read {}:".format(phmm_file, read.Name), file=sys.stderr)
        sys.exit(1)

    return list(_parse_hmmscan_tab(dout.decode().split("\n"), reads))



# -*- coding: utf-8 -*-

from six.moves import reduce
from numpy.random import random
from pysam import FastxFile
from math import log
import sys
from pychopper.common_structures import Seq

""" Utilities manipulating biological sequences and formats. Extensions to biopython functionality.
"""

# Reverse complements of bases, taken from dragonet:
comp = {
    'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'X': 'X', 'N': 'N',
    'a': 't', 't': 'a', 'c': 'g', 'g': 'c', 'x': 'x', 'n': 'n',
    '-': '-'
}


def base_complement(k):
    """ Return complement of base.

    Performs the subsitutions: A<=>T, C<=>G, X=>X for both upper and lower
    case. The return value is identical to the argument for all other values.

    :param k: A base.
    :returns: Complement of base.
    :rtype: str

    """
    try:
        return comp[k]
    except KeyError:
        sys.stderr.write(
            "WARNING: No reverse complement for {} found, returning argument.".format(k))
        return k


def reverse_complement(seq):
    """ Return reverse complement of a string (base) sequence.

    :param seq: Input sequence.
    :returns: Reverse complement of input sequence.
    :rtype: str

    """
    if len(seq) == 0:
        return seq
    return reduce(lambda x, y: x + y, map(base_complement, seq[::-1]))


def readfq(fastq, sample=None, min_qual=0, rfq_sup={}):  # this is a generator function
    """Read fastx files.

    This is a generator function that yields sequtils.Seq objects.
    Optionally filter by a minimum mean quality (min_qual).
    Optionally subsample the fastx file using sample (0.0 - 1.0)
    """
    sup = ("out_fq" in rfq_sup) and (rfq_sup["out_fq"] is not None)
    tsup = "total" in rfq_sup
    if sup:
        fh = open(rfq_sup["out_fq"], "w")

    with FastxFile(fastq) as fqin:
        for fx in fqin:
            if sample is None or (random() < sample):
                if tsup:
                    rfq_sup["total"] += 1
                probs = [10 ** (q / -10) for q in fx.get_quality_array()]
                if (-10 * log(sum(probs) / len(probs), 10)) >= min_qual:
                    if tsup:
                        rfq_sup["pass"] += 1
                    yield Seq(
                        Id=fx.name,
                        Name=f"{fx.name} {fx.comment}" if fx.comment else fx.name,
                        Seq=fx.sequence, Qual=fx.quality, Umi=None)
                else:
                    if sup:
                        fh.write(f"{fx}\n")
    if sup:
        fh.flush()
        fh.close()


def writefq(r, fh):
    "Write read to fastq file"
    q = r.Qual
    if q is None:
        q = "!" * len(r.Seq)
    fh.write("@{}\n{}\n+\n{}\n".format(r.Name, r.Seq, q))


def revcomp_seq(seq):
    """ Reverse complement sequence record """
    qual = seq.Qual
    if qual is not None:
        qual = qual[::-1]
    return Seq(seq.Id, seq.Name, reverse_complement(seq.Seq), qual, seq.Umi)


def get_runid(desc):
    """ Parse out runid from sequence description. """
    tmp = [t for t in desc.split(" ") if t.startswith("runid")]
    if len(tmp) != 1:
        return "NA"
    return tmp[0].rsplit("=", 1)[1]


def record_size(read, in_format='fastq'):
    """ Calculate record size. """
    if read.Qual is None:
        in_format = 'fasta'
    dl = len(read.Name)
    sl = len(read.Seq)
    if in_format == 'fastq':
        bl = dl + 2 * sl + 6
    elif in_format == 'fasta':
        bl = dl + sl + 3
    else:
        raise Exception("Unkonwn format!")
    return bl


def get_primers(primers):
    """Load primers from fasta file"""
    all_primers = {}
    with FastxFile(primers) as fh:
        for primer in fh:
            all_primers[primer.name] = primer.sequence
            all_primers['-' + primer.name] = reverse_complement(primer.sequence)
    return all_primers

def errs_tab(n):
    """Generate list of error rates for qualities less than equal than n."""
    return [10**(q / -10) for q in range(n + 1)]


def mean_qual(quals, qround=False, tab=errs_tab(128)):
    """Calculate average basecall quality of a read.
    Receive the ascii quality scores of a read and return the average quality for that read
    First convert Phred scores to probabilities,
    calculate average error probability
    convert average back to Phred scale
    """
    if quals:
        mq = -10 * log(sum([tab[ord(q) - 33] for q in quals]) / len(quals), 10)
        if qround:
            return round(mq)
        else:
            return mq
    else:
        return 0.0



# -*- coding: utf-8 -*-

from collections import namedtuple

Hit = namedtuple('Hit', 'Ref RefStart RefEnd Query QueryStart QueryEnd Score')
Seq = namedtuple('Seq', 'Id Name Seq Qual Umi')
Segment = namedtuple('Segment', 'Left Start End Right Strand Len')



# -*- coding: utf-8 -*-

import six
import numpy as np
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


class Report:

    # Maybe it would be a good idea to convert these utility methods to object
    # oriented matplotlib style.

    def __init__(self, pdf):
        """Class for plotting utilities on the top of matplotlib. Plots are saved in the specified file through the PDF backend.

        :param self: object.
        :param pdf: Output pdf.
        :returns: The report object.
        :rtype: Report

        """
        self.pdf = pdf
        self.plt = plt
        self.pages = PdfPages(pdf)

    def _set_properties_and_close(self, fig, title, xlab, ylab):
        """Utility method to set title, axis labels and close the figure.

        :param self: object.
        :param fig: The current figure.
        :param title: Figure title.
        :param xlab: X axis label.
        :param ylab: Y axis label.
        :returns: None
        :rtype: object
        """
        plt.xlabel(xlab)
        plt.ylabel(ylab)
        plt.title(title)
        self.pages.savefig(fig)
        plt.close(fig)

    def save_close(self):
        """Utility method to save and close figure."""
        self.pages.savefig()
        plt.close()
        plt.clf()

    def plot_arrays(self, data_map, title="", xlab="", ylab="", marker='.', legend_loc='best', legend=True, vlines=None, vlcolor='green', vlwitdh=0.5):
        """Plot multiple pairs of data arrays.

        :param self: object.
        :param data_map: A dictionary with labels as keys and tupples of data arrays (x,y) as values.
        :param title: Figure title.
        :param xlab: X axis label.
        :param ylab: Y axis label.
        :param marker: Marker passed to the plot function.
        :param legend_loc: Location of legend.
        :param legend: Plot legend if True
        :param vlines: Dictionary with labels and positions of vertical lines to draw.
        :param vlcolor: Color of vertical lines drawn.
        :param vlwidth: Width of vertical lines drawn.
        :returns: None
        :rtype: object
        """
        fig = plt.figure()

        for label, data_arrays in six.iteritems(data_map):
            plt.plot(data_arrays[0], data_arrays[1], marker, label=label)
        if vlines is not None:
            for label, pos in six.iteritems(vlines):
                plt.axvline(x=pos, label=label, color=vlcolor, lw=vlwitdh)
        if legend:
            plt.legend(loc=legend_loc)
        self._set_properties_and_close(fig, title, xlab, ylab)

    def plot_histograms(self, data_map, title="", xlab="", ylab="", bins=50, alpha=0.7, legend_loc='best', legend=True, vlines=None):
        """Plot histograms of multiple data arrays.

        :param self: object.
        :param data_map: A dictionary with labels as keys and data arrays as values.
        :param title: Figure title.
        :param xlab: X axis label.
        :param ylab: Y axis label.
        :param bins: Number of bins.
        :param alpha: Transparency value for histograms.
        :param legend_loc: Location of legend.
        :param legend: Plot legend if True.
        :param vlines: Dictionary with labels and positions of vertical lines to draw.
        :returns: None
        :rtype: object
        """
        fig = plt.figure()

        for label, data in six.iteritems(data_map):
            if len(data) > 0:
                plt.hist(data, bins=bins, label=label, alpha=alpha)
        if vlines is not None:
            for label, pos in six.iteritems(vlines):
                plt.axvline(x=pos, label=label)
        if legend:
            plt.legend(loc=legend_loc)

        self._set_properties_and_close(fig, title, xlab, ylab)

    def plot_bars_simple(self, data_map, title="", xlab="", ylab="", alpha=0.6, xticks_rotation=0, auto_limit=False):
        """Plot simple bar chart from input dictionary.

        :param self: object.
        :param data_map: A dictionary with labels as keys and data as values.
        :param title: Figure title.
        :param xlab: X axis label.
        :param ylab: Y axis label.
        :param alpha: Alpha value.
        :param xticks_rotation: Rotation value for x tick labels.
        :param auto_limit: Set y axis limits automatically.
        :returns: None
        :rtype: object
        """
        fig = plt.figure()

        labels = list(data_map.keys())
        data = list(data_map.values())
        positions = np.arange(len(labels))
        plt.bar(positions, data, align='center', alpha=alpha)
        plt.xticks(positions, labels, rotation=xticks_rotation)

        if auto_limit:
            low, high = min(data), max(data)
            plt.ylim([(low - 0.5 * (high - low)), (high + 0.5 * (high - low))])

        self._set_properties_and_close(fig, title, xlab, ylab)

    def close(self):
        """Close PDF backend. Do not forget to call this at the end of your script or your output will be damaged!

        :param self: object
        :returns: None
        :rtype: object
        """
        self.pages.close()




from collections import OrderedDict
import subprocess as sp
from itertools import islice, chain
import numpy as np
import sys


def parse_config_string(s):
    res = OrderedDict()
    s = s.strip()
    for token in s.split("|"):
        token = token.strip().split(":")
        if len(token) != 2:
            raise Exception("Invalid token: " + token)
        strand = token[0].strip()
        if strand not in ('+', '-'):
            raise Exception("Invalid strand: " + strand)
        token = token[1].strip().split(",")
        res[(token[0].strip(), token[1].strip())] = strand
    return res


def batch(iterable, size):
    sourceiter = iter(iterable)
    while True:
        batchiter = islice(sourceiter, size)
        try:
            yield list(chain([next(batchiter)], batchiter))
        except StopIteration:
            return


def hit2bed(hit, read):
    # Hit = namedtuple('Hit', 'Ref RefStart RefEnd Query QueryStart QueryEnd
    # Score')
    max_q, q = 100, None
    if hit.Score == 0:
        q = max_q
    elif hit.Score > 1:
        q = 0
    else:
        q = -10 * np.log10(hit.Score)
    name = hit.Query
    strand = "+"
    if hit.Query[0] == "-":
        name = hit.Query[1:]
        strand = "-"
    bed_line = "%s\t%d\t%d\t%s\t%d\t%s" % (read.Name, hit.RefStart, hit.RefEnd, name, q, strand)
    return bed_line


def count_fastq_records(fname, size=128000000, opener=open):
    fh = opener(fname, "r")
    count = 0
    while True:
        b = fh.read(size)
        if not b:
            break
        count += b.count("\n+\n")
    fh.close()
    return count


def check_command(cmd):
    if sp.call(cmd, shell=True) != 0:
        sys.stderr.write("Required command {} not found in the path!\n".format(cmd.split(" ")[0]))


def check_min_hmmer_version(major, minor):
    ver_line = sp.check_output("nhmmscan -h | grep '^# HMMER'", shell=True).decode()
    tmp = ver_line.split("HMMER")[1].strip().split(" ")[0].split(".")
    f_major, f_minor = int(tmp[0]), int(tmp[1])
    if f_major < major:
        raise Exception("Insufficient major HMMER version: " + str(f_major))
    if f_minor < minor:
        raise Exception("Insufficient minor HMMER version: " + str(f_minor))



#!/usr/bin/env python

"""Helper script to generate cmd_line.rst file for all scripts in bin which
have and parser object defined in their global scope - taken from tang

"""
from __future__ import print_function
import sys
import os
import imp

scripts_rel = 'scripts'
attr_name = 'parser'
blacklist = ['__init__.py']

location = os.path.join(os.path.dirname(os.path.abspath(__file__)),'..')
scripts_abs = os.path.join(location, scripts_rel)
scripts = sorted(filter(lambda s: s[0] != '.' and s not in blacklist, os.listdir(scripts_abs)))

sys.stderr.write("Found following scripts:\n{}\n{}\n{}\n".format(
    location, scripts_abs, scripts
))


print ("""
.. _command_line_tools:

Command line tools
==================
""")

for script in scripts:
    script_name, script_ext = os.path.splitext(script)
    if script_ext == '.pyc':
        continue
    
    try:
        mod_name = '{}.{}'.format(scripts_rel, script_name)
        #mod = __import__(mod_name, globals(), locals(), [attr_name])
        mod = imp.load_source(script_name, os.path.join(scripts_abs, script))
        script = script.replace('.py', '')

        print ('.. _{}:\n\n{}\n{}'.format(script, script, '-'*len(script)))
        if hasattr(mod, attr_name):
            print ("""
.. argparse::
   :ref: {}.{}
   :prog: {}
""".format(mod_name, attr_name, script_name))
        else:
            print('No documentation available')

    except Exception as e:
        # Wha' yer' gonna do?
        sys.stderr.write('Error making docs for {}:\n{}\n'.format(script_name, e))
        pass



#!/usr/bin/env python

import sys
import pandas as pd
from sklearn import linear_model
import numpy as np
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns

F1, F2, TSV, PDF = sys.argv[1:]
df = pd.read_csv(TSV, sep="\t")
print(df.describe())

regr = linear_model.LinearRegression()
X = df[F1].values.reshape(-1, 1)
y = df[F2].values.reshape(-1, 1)
regr.fit(X, y)
a = float(regr.coef_[0])
b = float(regr.intercept_)
eq = "y = {:.3f}x + {:.3f}".format(a, b)
print(eq)

pages = PdfPages(PDF)

g = sns.lmplot(F1, F2, df, fit_reg=True, aspect=1.5, ci=95, scatter_kws={"s": 1})
props = dict(boxstyle='round', alpha=0.5, color=sns.color_palette()[0])
textstr = '$' + eq + '$'
g.ax.text(0.7, 0.05, textstr, transform=g.ax.transAxes, fontsize=14, bbox=props)

pages.savefig()
plt.close()
pages.close()



# -*- coding: utf-8 -*-

import edlib

from pychopper import seq_utils as seu
from pychopper.common_structures import Hit
from pychopper import utils
from pychopper.parasail_backend import refine_locations


def find_locations(reads, all_primers, max_ed, pool, min_batch):
    "Find alignment hits of all primers in all reads using the edlib/parasail backend"
    for batch in utils.batch(reads, min_batch):
        for res in pool.map(_find_locations_single, zip(batch, [(all_primers, max_ed)] * len(batch))):
            try:
                yield res
            except StopIteration:
                return


def find_umi_single(params):
    "Find UMI in a single reads using the edlib/parasail backend"
    read = params[0]
    max_ed = params[1]

    pattern_list = [
        (
            "TTTVVVVTTVVVVTTVVVVTTVVVVTTT",
            "V",
            [("V", "A"), ("V", "G"), ("V", "C")],
            True
        ),
        (
            "AAABBBBAABBBBAABBBBAABBBBAAA",
            "B",
            [("B", "T"), ("B", "G"), ("B", "C")],
            False
        )
    ]

    best_ed = None
    best_pattern = None
    best_result = None
    for pattern, wildcard, equalities, forward in pattern_list:
        result = edlib.align(pattern,
                             read,
                             task="path",
                             mode="HW",
                             k=max_ed,
                             additionalEqualities=equalities)
        if result['editDistance'] == -1:
            continue
        if not best_ed or result['editDistance'] < best_ed:
            best_ed = result['editDistance']
            best_pattern = (pattern, wildcard, equalities, forward)
            best_result = result

    if best_ed is None:
        return None, None

    # Extract UMI
    pattern, wildcard, equalities, forward = best_pattern
    ed = best_result["editDistance"]

    locs = best_result["locations"][0]
    umi = read[locs[0]:locs[1]+1]
    if not forward:
        umi = seu.reverse_complement(umi)
    # Do not assign UMIs where UMI probe has aligned to the 'N' spacer.
    if 'N' in umi:
        return None, None
    return umi, ed


def _find_locations_single(params):
    "Find alignment hits of all primers in a single reads using the edlib/parasail backend"
    read = params[0]
    all_primers, max_ed = params[1]
    all_locations = []
    for primer_acc, primer_seq in all_primers.items():
        primer_max_ed = int(max_ed * len(primer_seq))
        result = edlib.align(primer_seq, read.Seq,
                             mode="HW", task="locations", k=primer_max_ed)
        ed = result["editDistance"]
        locations = result["locations"]
        if locations:
            # all_locations[primer_acc] = []
            for refstart, refend in locations:
                refend += 1
                # ('Hit', 'Ref RefStart RefEnd Query QueryStart QueryEnd Score')
                hit = Hit(read.Name, refstart, refend, primer_acc, 0,
                          len(primer_seq),  ed / len(primer_seq))
                all_locations.append(hit)
    refined_locations = refine_locations(read, all_primers, all_locations)
    return refined_locations



# -*- coding: utf-8 -*-

import sys
import numpy as np
from pychopper import seq_utils as seu
from pychopper import hmmer_backend, edlib_backend
from pychopper.common_structures import Segment, Seq
from pychopper.alignment_hits import process_hits


def _build_segments(hits, config):
    "Build tuple of segments"
    segments = []
    if len(hits) == 0:
        return tuple(segments)
    for s0,s1 in zip(hits, hits[1:]):
        strand = None
        seg_len = 0
        c = (s0.Query, s1.Query)
        if c in config:
            strand = config[c]
            seg_len = s1.RefStart - s0.RefEnd
        segments.append(Segment(s0.RefStart, s0.RefEnd, s1.RefStart, s1.RefEnd, strand, seg_len))
    return tuple(segments)


def analyse_hits(hits, config):
    """ Segment reads based on alignment hits using dynamic programming.
        The algorithm is based on the rule that each primer alignment hit can be used only once.
        Hence if a segment is included, the next one has to be excluded.
    """
    segments = _build_segments(hits, config)
    nr = len(segments)

    # No hits, return no segments:
    if nr == 0:
        return (), (), 0

    # Initialize DP matrix and traceback dictionary:
    M = np.zeros((2, nr), dtype=int)
    B = dict()

    # Initialize entry for the first segment:
    M[1, 0] = segments[0].Len

    # Fill in DP matrix:
    for j in range(1, nr):
        # First row holds excluded segments.
        # The can transition from eiter exluded or included segments:
        M[0, j] = M[0, j - 1]
        B[0, j] = (0, j - 1)
        if M[1, j - 1] > M[0, j - 1]:
            M[0, j] = M[1, j - 1]
            B[0, j] = (1, j - 1) 
        # Included segments can only transition from previously excluded segments:
        M[1, j] = M[0, j - 1] + segments[j].Len
        B[1, j] = (0, j - 1)

    tlen = np.argmax(M[:, nr - 1])
    valid_segments = []

    # Traceback and build solution:
    p = (tlen, nr - 1)
    while True:
        if p[0] == 1:
            s = segments[p[1]]
            # Filter out invalid segments which can be part of the
            # solution:
            if s.Len > 0:
                valid_segments.append(s)
        if p[1] == 0:
            break
        p = B[p]

    return tuple(valid_segments), hits, tlen


def segments_to_reads(read, segments, keep_primers, bam_tags, detect_umis):
    "Convert segments to output reads with annotation"
    for s in segments:
        Start = s.Start
        End = s.End
        if keep_primers:
            Start = s.Left
            End = s.Right

        # Format FASTQ name and comment
        sr_id = "{}:{}|".format(Start, End)
        if bam_tags:
            try:
                name, comment = read.Name.split(" ", 1)
            except ValueError:
                name = read.Name
                comment = ""
            sr_name = sr_id + name + " CO:Z:" + comment + "\tTS:A:{}".format(s.Strand)
        else:
            sr_name = sr_id + read.Name + " strand=" + s.Strand

        if len(segments) > 1:
            sr_name += " rescue=1"

        umi = None
        if detect_umis:
            max_umi_ed = 3
            # Detect UMIs
            # Get adapters for UMI search
            padding = 40
            p1_from = max(0, s.Left - padding)
            p1_to = min(len(read.Seq), s.Start + padding)
            p_1 = read.Seq[p1_from:p1_to]
            p2_from = max(0, s.End - padding)
            p2_to = min(len(read.Seq), s.Right + padding)
            p_2 = read.Seq[p2_from:p2_to]

            # Create a single probe containing adapter-surrounding sequences.
            # to be used in a single UMI search. Separate with 'N' spacer to prevent
            # overlapping hits
            umi_scan_seq = p_1 + 'NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN' + p_2

            # If the read is small and the adapter regions + padding overlap,
            # then use only the whole read for the UMI search
            if p1_to >= p2_from:
                umi_scan_seq = read.Seq

            umi, _ = edlib_backend.find_umi_single(
                [umi_scan_seq, max_umi_ed])
            if bam_tags:
                sr_name += "\tRX:Z:{}".format(umi)
            else:
                sr_name += " umi={}".format(umi)

        sr_seq = read.Seq[Start:End]
        sr = Seq(sr_id + read.Id, sr_name, sr_seq, read.Qual[Start:End] if read.Qual is not None else None, umi)
        if s.Strand == '-':
            sr = seu.revcomp_seq(sr)
        yield sr


def chopper_phmm(reads, phmm_file, config, cutoff, threads, pool, min_batch):
    "Segment using the profile HMM backend"
    batch_hits = hmmer_backend.find_locations(reads, phmm_file, E=cutoff, pool=pool, min_batch=min_batch)
    for i, hits in enumerate(batch_hits):
        hits = process_hits(hits, cutoff)
        yield reads[i], analyse_hits(hits, config)


def chopper_edlib(reads, primers, config, max_ed, cutoff, pool, min_batch):
    "Segment using the edlib/parasail backend"
    batch_hits = edlib_backend.find_locations(reads, primers, max_ed=max_ed, pool=pool, min_batch=min_batch)
    for i, hits in enumerate(batch_hits):
        hits = process_hits(hits, cutoff)
        yield reads[i], analyse_hits(hits, config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import io
import sys
import numpy as np
import pandas as pd
from collections import OrderedDict, defaultdict
import concurrent.futures
import tqdm
import gzip

from pychopper import seq_utils as seu
from pychopper import utils
from pychopper import chopper, report
import pychopper.phmm_data as phmm_data
import pychopper.primer_data as primer_data


def _new_stats():
    "Initialize a new statistic dictionary"
    st = OrderedDict()
    st["Classification"] = OrderedDict([('Primers_found', 0), ('Rescue', 0), ('Unusable', 0)])
    st["Strand"] = OrderedDict([('+', 0), ('-', 0)])
    st["RescueStrand"] = OrderedDict([('+', 0), ('-', 0)])
    st["RescueSegmentNr"] = defaultdict(int)
    st["RescueHitNr"] = defaultdict(int)
    st["UnclassHitNr"] = defaultdict(int)
    st["Unusable"] = defaultdict(int)
    st["Hits"] = defaultdict(int)
    st["LenFail"] = 0
    st["QcFail"] = 0
    st["PassReads"] = 0
    st["Umi_detected"] = 0
    st["Umi_detected_final"] = 0
    return st


def _update_stats(st, d_fh, segments, hits, usable_len, read):
    "Update stats dictionary with properties of a read"
    st["PassReads"] += 1
    if len(hits) > 0:
        h = ",".join([x.Query for x in hits])
        st["Hits"][h] += 1
    if len(segments) == 0:
        st["Classification"]["Unusable"] += 1
        st["UnclassHitNr"][len(hits)] += 1
        if d_fh is not None:
            d_fh.write("{}\t{}\t0\t{}\t{}\t{}\n".format(read.Name, len(read.Seq), -1, -1, "."))
    elif len(segments) == 1:
        st["Classification"]["Primers_found"] += 1
        st["Strand"][segments[0].Strand] += 1
        st["Unusable"][int(segments[0].Len / len(read.Seq) * 100)] += 1
        if d_fh is not None:
            rs = segments[0]
            d_fh.write("{}\t{}\t1\t{}\t{}\t{}\n".format(read.Name, len(read.Seq), rs.Start, rs.End, rs.Strand))
    else:
        for rs in segments:
            st["Classification"]["Rescue"] += 1
            st["RescueStrand"][rs.Strand] += 1
            st["RescueHitNr"][len(hits)] += 1
            if d_fh is not None:
                d_fh.write("{}\t{}\t{}\t{}\t{}\t{}\n".format(read.Name, len(read.Seq), len(segments), rs.Start, rs.End, rs.Strand))
        st["Unusable"][len(read.Seq) - int(sum([s.Len for s in segments]))] += 1
        st["RescueSegmentNr"][len(segments)] += 1


def _process_stats(st):
    "Convert stats dictionary into a data frame"
    res = OrderedDict([("Category", []), ("Name", []), ("Value", [])])
    res["Category"] += ["ReadStats"]
    res["Name"] += ["PassReads"]
    res["Value"] += [st["PassReads"]]
    res["Category"] += ["ReadStats"]
    res["Name"] += ["LenFail"]
    res["Value"] += [st["LenFail"]]
    res["Category"] += ["ReadStats"]
    res["Name"] += ["QcFail"]
    res["Value"] += [st["QcFail"]]
    for k, v in st['Classification'].items():
        res["Category"] += ["Classification"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in st['Strand'].items():
        res["Category"] += ["Strand"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in st['RescueStrand'].items():
        res["Category"] += ["RescueStrand"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in sorted(st['UnclassHitNr'].items(), key=lambda x: x[0]):
        res["Category"] += ["UnclassHitNr"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in sorted(st['RescueHitNr'].items(), key=lambda x: x[0]):
        res["Category"] += ["RescueHitNr"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in sorted(st['RescueSegmentNr'].items(), key=lambda x: x[0]):
        res["Category"] += ["RescueSegmentNr"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in sorted(st['Unusable'].items(), key=lambda x: x[0]):
        res["Category"] += ["Unusable"]
        res["Name"] += [k]
        res["Value"] += [v]
    for k, v in sorted(st['Hits'].items(), key=lambda x: x[1], reverse=True):
        res["Category"] += ["Hits"]
        res["Name"] += [k]
        res["Value"] += [v]

    res["Category"] += ["ReadStats"]
    res["Name"] += ["Umi_detected"]
    res["Value"] += [st["Umi_detected"]]
    res["Category"] += ["ReadStats"]
    res["Name"] += ["Umi_detected_final"]
    res["Value"] += [st["Umi_detected_final"]]
    res = pd.DataFrame(res)
    return res


def _detect_anomalies(st, config):
    raw_anom = []
    total = st["PassReads"]
    for k, v in sorted(st['Hits'].items(), key=lambda x: x[1], reverse=True):
        x = tuple(k.split(","))
        if len(x) != 2:
            raw_anom.append((k, v))
        elif x not in config:
            raw_anom.append((k, v))
    anom = []
    for tmp in raw_anom:
        pc = tmp[1] * 100.0 / total
        if pc >= 3.0:
            anom.append((tmp[0], tmp[1], pc))
    if len(anom) > 0:
        sys.stderr.write("Detected {} potential artefactual primer configurations:\n".format(len(anom)))
        ml = max([len(t[0]) for t in anom])
        mn = max([len(str(t[1])) for t in anom])
        sys.stderr.write(("{:" + str(ml) + "}\t{:" + str(mn) + "}\t{}\n").format("Configuration", "NrReads", "PercentReads"))
        for x in anom:
            fs = "{:" + str(ml) + "}\t{:" + str(mn) + "}\t{:.2f}%\n"
            sys.stderr.write(fs.format(x[0], str(x[1]), x[2]))


def _plot_pd_bars(df, title, report, alpha=0.7, xrot=0, ann=False):
    "Generate bar plot from data frame"
    df = df.drop("Category", axis=1)
    df = df.set_index("Name")
    report.plt.clf()
    if sum(df.Value) > 0:
        ax = df.plot(kind='bar', align='center', alpha=alpha, rot=xrot)
        report.plt.title(title)
        if ann:
            y_offset = 1
            for rect in ax.patches:
                height = rect.get_height()
                ax.text(rect.get_x() + rect.get_width() / 2., height + y_offset, "{:.0f}".format(height), ha='center', va='bottom', rotation=0)
                ax.text(rect.get_x() + rect.get_width() / 2., 0.5 * height - 2 * y_offset, "{:.1f}%".format(100 * height / sum(df.Value)), ha='center', va='bottom', rotation=0)
        report.save_close()


def _plot_pd_line(df, title, report, alpha=0.7, xrot=0, vline=None):
    "Generate line plot from a data frame"
    df = df.drop("Category", axis=1)
    df = df.set_index("Name")
    report.plt.clf()
    if sum(df.Value) > 0:
        df.plot(kind='line', rot=xrot)
        if vline is not None:
            report.plt.axvline(x=vline, color="red")
        report.plt.title(title)
        report.save_close()


def _plot_stats(st, pdf, q, q_bak, detect_umi):
    "Generate plots and save to report PDF"
    R = report.Report(pdf)
    rs = st.loc[st.Category == "Classification", ]
    _plot_pd_bars(rs.copy(), "Classification of output reads", R, ann=True)
    found, rescue, unusable = float(rs.loc[rs.Name == "Primers_found", ].Value), float(rs.loc[rs.Name == "Rescue", ].Value), float(rs.loc[rs.Name == "Unusable", ].Value)
    rs_stats = st.loc[st.Category == "ReadStats", ]
    umi_detected_final = float(rs_stats.loc[rs_stats.Name == "Umi_detected_final", 'Value'].iloc[0])
    total = found + rescue + unusable
    found = found / total * 100
    rescue = rescue / total * 100
    unusable = unusable / total * 100
    umi_detected_final = umi_detected_final / total * 100
    sys.stderr.write("-----------------------------------\n")
    if detect_umi:
        sys.stderr.write("Reads with two primers:\t{:.2f}% (with UMI {:.2f}%)\nRescued reads:\t\t{:.2f}%\nUnusable reads:\t\t{:.2f}%\n".format(found, umi_detected_final, rescue, unusable))
    else:
        sys.stderr.write("Reads with two primers:\t{:.2f}%\nRescued reads:\t\t{:.2f}%\nUnusable reads:\t\t{:.2f}%\n".format(found, rescue, unusable))
    sys.stderr.write("-----------------------------------\n")
    _plot_pd_bars(st.loc[st.Category == "Strand", ].copy(), "Strand of oriented reads", R, ann=True)
    _plot_pd_bars(st.loc[st.Category == "RescueStrand", ].copy(), "Strand of rescued reads", R, ann=True)
    _plot_pd_bars(st.loc[st.Category == "UnclassHitNr", ].copy(), "Number of hits in unclassified reads", R)
    _plot_pd_bars(st.loc[st.Category == "RescueHitNr", ].copy(), "Number of hits in rescued reads", R)
    _plot_pd_bars(st.loc[st.Category == "RescueSegmentNr", ].copy(), "Number of usable segments per rescued read", R)
    if q_bak is None:
        _plot_pd_line(st.loc[st.Category == "AutotuneSample", ].copy(), "Usable bases as function of cutoff(q). Best q={:.4g}".format(q), R, vline=q)
    udf = st.loc[st.Category == "Unusable", ].copy()
    udf.Name = np.log10(1.0 + np.array(udf.Name, dtype=float))
    _plot_pd_line(udf, "Log10 length distribution of trimmed away sequences.", R)
    R.close()


def _opener(filename, mode, encoding='utf8'):
    if filename == '-':
        sys.stderr.write("Reading from stdin\n")
        #return open(sys.stdin.buff, mode, encoding=encoding)
        return io.TextIOWrapper(sys.stdin.buffer, encoding=encoding)
    elif filename.endswith('.gz'):
        return gzip.open(filename, mode + 't', encoding=encoding)
    else:
        return open(filename, mode, encoding=encoding)


def main():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description='Tool to identify, orient and rescue full-length cDNA reads.')
    parser.add_argument(
        '-b', metavar='primers', type=str, default=None, help="Primers fasta.",
        required=False)
    parser.add_argument(
        '-g', metavar='phmm_file', type=str, default=None,
        help="File with custom profile HMMs (None).", required=False)
    parser.add_argument(
        '-c', metavar='config_file', type=str, default=None,
        help="File to specify primer configurations for each direction (None).")
    parser.add_argument(
        '-k', type=str, default="PCS109",
        help="Use primer sequences from this kit (PCS109).",
        choices=['PCS109', 'PCS110', 'PCS111', 'LSK114']
    )
    parser.add_argument(
        '-q', metavar='cutoff', type=float, default=None,
        help="Cutoff parameter (autotuned).")
    parser.add_argument(
        '-Q', metavar='min_qual', type=float, default=7.0,
        help="Minimum mean base quality (7.0).")
    parser.add_argument(
        '-z', metavar='min_len', type=int, default=50,
        help="Minimum segment length (50).")
    parser.add_argument(
        '-r', metavar='report_pdf', type=str,
        default="pychopper.pdf",
        help="Report PDF (pychopper.pdf).")
    parser.add_argument(
        '-u', metavar='unclass_output', type=str, default=None,
        help="Write unclassified reads to this file.")
    parser.add_argument(
        '-l', metavar='len_fail_output', type=str, default=None,
        help="Write fragments failing the length filter in this file.")
    parser.add_argument(
        '-w', metavar='rescue_output', type=str, default=None,
        help="Write rescued reads to this file.")
    parser.add_argument(
        '-S', metavar='stats_output', type=str,
        default="pychopper.tsv",
        help="Write statistics to this file.")
    parser.add_argument(
        '-K', metavar='qc_fail_output', type=str, default=None,
        help="Write reads failing mean quality filter to this file.")
    parser.add_argument(
        '-Y', metavar='autotune_nr', type=float, default=10000,
        help="Approximate number of reads used for tuning the cutoff parameter (10000).")
    parser.add_argument(
        '-L', metavar='autotune_samples', type=int, default=30,
        help="Number of samples taken when tuning cutoff parameter (30).")
    parser.add_argument(
        '-A', metavar='scores_output', type=str, default=None,
        help="Write alignment scores to this BED file.")
    parser.add_argument(
        '-m', metavar='method', type=str, default="phmm",
        help="Detection method: phmm or edlib (phmm).")
    parser.add_argument(
        '-x', metavar='rescue', type=str, default=None,
        help="Protocol-specific read rescue: DCS109 (None).")
    parser.add_argument(
        '-p', action='store_true', default=False,
        help="Keep primers, but trim the rest.")
    parser.add_argument(
        '-t', metavar='threads', type=int, default=8,
        help="Number of threads to use (8).")
    parser.add_argument(
        '-B', metavar='batch_size', type=int, default=10000,
        help="Maximum number of reads processed in each batch (10000).")
    parser.add_argument(
        '-D', metavar='read stats', type=str, default=None,
        help="Tab separated file with per-read stats (None).")
    parser.add_argument(
        '-y', action='store_true', default=False,
        help="Output FASTQ comment as BAM tags. Use with minimap2 -y to pass UMI and additional info into BAM file.")
    parser.add_argument(
        '-U', action='store_true', default=False,
        help="Detect UMIs")

    parser.add_argument('input_fastx', metavar='input_fastx', type=str,
                        help="Input file.")
    parser.add_argument('output_fastx', metavar='output_fastx', nargs="?",
                        type=str, default="-", help="Output file.")

    args = parser.parse_args()

    if args.m == "phmm":
        utils.check_command("nhmmscan -h > /dev/null")
        utils.check_min_hmmer_version(3, 2)

    CONFIG = "+:SSP,-VNP|-:VNP,-SSP"
    if args.c is not None:
        CONFIG = open(args.c, "r").readline().strip()

    kits = {
        "PCS109": {
            "HMM": os.path.join(
                os.path.dirname(phmm_data.__file__), "cDNA_SSP_VNP.hmm"),
            "FAS": os.path.join(
                os.path.dirname(primer_data.__file__), "cDNA_SSP_VNP.fas"),
        },
        "PCS110": {
            "HMM": os.path.join(
                os.path.dirname(phmm_data.__file__), "PCS110_primers.hmm"),
            "FAS": os.path.join(
                os.path.dirname(primer_data.__file__), "PCS110_primers.fas")
        },
        "PCS111": {
            "HMM": os.path.join(
                os.path.dirname(phmm_data.__file__), "PCS110_primers.hmm"),
            "FAS": os.path.join(
                os.path.dirname(primer_data.__file__), "PCS111_primers.fas")},
        "LSK114": {
            "HMM": os.path.join(
                os.path.dirname(phmm_data.__file__), "cDNA_SSP_VNP.hmm"),
            "FAS": os.path.join(
                os.path.dirname(primer_data.__file__), "LSK114_primers.fas")}
    }

    if args.g is None:
        args.g = kits[args.k]["HMM"]
    elif args.m != 'phmm':
        sys.exit(
            'if using -g option, phmm backend should be used (-m phmm)'
        )

    if args.b is None:
        args.b = kits[args.k]["FAS"]
    elif args.m != 'edlib':
        sys.exit(
            'if using -b option, edlib backend should be used (-m edlib)'
        )

    if args.x is not None and args.x in ('DCS109'):
        if args.x == "DCS109":
            CONFIG = "-:VNP,-VNP"

    config = utils.parse_config_string(CONFIG)
    sys.stderr.write("Using kit: {}\n".format(args.b if args.b else args.k))
    sys.stderr.write("Configurations to consider: \"{}\"\n".format(CONFIG))

    out_fh = sys.stdout
    if args.output_fastx != '-':
        out_fh = open(args.output_fastx, "w")

    u_fh = None
    if args.u is not None:
        u_fh = open(args.u, "w")

    l_fh = None
    if args.l is not None:
        l_fh = open(args.l, "w")

    w_fh = None
    if args.w is not None:
        w_fh = open(args.w, "w")

    a_fh = None
    if args.A is not None:
        a_fh = open(args.A, "w")

    d_fh = None
    if args.D is not None:
        d_fh = open(args.D, "w")
        d_fh.write("Read\tLength\tStatus\tStart\tEnd\tStrand\n")

    st = _new_stats()
    input_size = None
    if args.input_fastx != "-":
        input_size = os.stat(args.input_fastx).st_size

    if args.q is None and args.Y <= 0:
        sys.stderr.write("Please specifiy either -q or -Y!")

    if args.m == "edlib":
        all_primers = seu.get_primers(args.b)

    if args.m == "phmm":
        def backend(x, pool, q=None, mb=None):
            return chopper.chopper_phmm(x, args.g, config, q, args.t, pool, mb)
    elif args.m == "edlib":
        def backend(x, pool, q=None, mb=None):
            return chopper.chopper_edlib(x, all_primers, config, q * 1.2, q,
                                         pool, mb)
    else:
        raise Exception("Invalid backend!")

    # Pick the -q maximizing the number of classified reads using grid search:
    #nr_records = None
    nr_records = None
    tune_df = None
    q_bak = args.q

    if args.q is None:
        nr_cutoffs = args.L
        cutoffs = np.linspace(0.0, 1.0, num=nr_cutoffs)
        cutoffs = cutoffs / cutoffs[-1]
        if args.m == "phmm":
            cutoffs = np.linspace(10 ** -5, 5.0, num=nr_cutoffs)
        class_reads = []
        class_readLens = []
        nr_records = utils.count_fastq_records(args.input_fastx,
                                               opener=_opener)
        opt_batch = int(nr_records / args.t)
        if opt_batch < args.B:
            args.B = opt_batch
        target_prop = 1.0
        if nr_records > args.Y:
            target_prop = args.Y / float(nr_records)
        if target_prop > 1.0:
            target_prop = 1.0
        sys.stderr.write("Counting fastq records in input file: {}\n".format(
            args.input_fastx))
        sys.stderr.write(
            "Total fastq records in input file: {}\n".format(nr_records))
        read_sample = list(
            seu.readfq(args.input_fastx, sample=target_prop, min_qual=args.Q))
        sys.stderr.write(
            "Tuning the cutoff parameter (q) on {} sampled reads ({:.1f}%) passing quality filters (Q >= {}).\n".format(
                len(read_sample), target_prop * 100.0, args.Q))
        sys.stderr.write("Optimizing over {} cutoff values.\n".format(args.L))
        for qv in tqdm.tqdm(cutoffs):
            clsLen = 0
            cls = 0
            with concurrent.futures.ProcessPoolExecutor(max_workers=args.t) as executor:
                for batch in utils.batch(read_sample, int((len(read_sample)))):
                    for read, (segments, hits, usable_len) \
                            in backend(batch, executor, qv, max(1000, int((len(read_sample)) / args.t))):
                        flt = list([x.Len for x in segments if x.Len > 0])
                        if len(flt) == 1:
                            clsLen += sum(flt)
                            cls += 1
            class_reads.append(cls)
            class_readLens.append(clsLen)
        best_qi = np.argmax(class_readLens)
        args.q = cutoffs[best_qi]
        tune_df = OrderedDict([("Category", []), ("Name", []), ("Value", [])])
        for i, c in enumerate(cutoffs):
            tune_df["Category"] += ["AutotuneSample"]
            tune_df["Name"] += [c]
            tune_df["Value"] += [class_reads[i]]

        tune_df["Category"] += ["Parameter"]
        tune_df["Name"] += ["Cutoff(q)"]
        tune_df["Value"] += [args.q]
        if best_qi == (len(class_reads) - 1):
            sys.stderr.write(
                "Best cuttoff value is at the edge of the search interval! Using tuned value is not safe! Please pick a q value manually and QC your data!\n")
        sys.stderr.write(
            "Best cutoff (q) value is {:.4g} with {:.0f}% of the reads classified.\n".format(
                args.q, class_reads[best_qi] * 100 / len(read_sample)))

    if nr_records is not None:
        input_size = nr_records
        if args.B > nr_records:
            args.B = nr_records
        if args.B == 0:
            args.B = 1
    sys.stderr.write(
        "Processing the whole dataset using a batch size of {}:\n".format(
            args.B))
    pbar = tqdm.tqdm(total=input_size)
    min_batch_size = max(int(args.B / args.t), 1)
    rfq_sup = {"out_fq": args.K, "pass": 0, "total": 0}
    with concurrent.futures.ProcessPoolExecutor(
            max_workers=args.t) as executor:
        for batch in utils.batch(
                seu.readfq(args.input_fastx, min_qual=args.Q, rfq_sup=rfq_sup), args.B):
            for read, (segments, hits, usable_len) in backend(batch, executor,
                                                              q=args.q,
                                                              mb=min_batch_size):
                if args.A is not None:
                    for h in hits:
                        a_fh.write(utils.hit2bed(h, read) + "\n")
                _update_stats(st, d_fh, segments, hits, usable_len, read)
                if args.u is not None and len(segments) == 0:
                    seu.writefq(read, u_fh)
                for trim_read in chopper.segments_to_reads(read, segments,
                                                           args.p, args.y, args.U):
                    if trim_read.Umi:
                        st["Umi_detected"] += 1
                    if len(trim_read.Seq) < args.z:
                        st["LenFail"] += 1
                        if args.l is not None:
                            seu.writefq(trim_read, l_fh)
                        continue
                    if len(segments) == 1:
                        if trim_read.Umi:
                            st["Umi_detected_final"] += 1
                        seu.writefq(trim_read, out_fh)
                    if args.w is not None and len(segments) > 1:
                        seu.writefq(trim_read, w_fh)
                #if nr_records is None:
                #    pbar.update(seu.record_size(read, 'fastq'))
                #else:
                pbar.update(1)
    pbar.close()
    sys.stderr.write("Finished processing file: {}\n".format(args.input_fastx))
    fail_nr = rfq_sup["total"] - rfq_sup["pass"]
    fail_pc = (fail_nr * 100 / rfq_sup["total"])
    st["QcFail"] = fail_nr
    sys.stderr.write(
        "Input reads failing mean quality filter (Q < {}): {} ({:.2f}%)\n".format(
            args.Q, fail_nr, fail_pc))
    sys.stderr.write(
        "Output fragments failing length filter (length < {}): {}\n".format(
            args.z, st["LenFail"]))

    # Save stats as TSV:
    stdf = None
    if args.S is not None or args.r is not None:
        stdf = _process_stats(st)
        if tune_df is not None:
            stdf = pd.concat([stdf, pd.DataFrame(tune_df)])

    _detect_anomalies(st, config)

    if args.S is not None:
        stdf.to_csv(args.S, sep="\t", index=False)

    for fh in (out_fh, u_fh, l_fh, w_fh, a_fh, d_fh):
        if fh is None:
            continue
        fh.flush()
        fh.close()

    if args.r is not None:
        _plot_stats(stdf, args.r, args.q, q_bak, args.U)


if __name__ == '__main__':
    main()

